from copy import deepcopy
import immutables
import pytest
from pathlib import Path

from skepticoin.humans import computer
from skepticoin.params import (
    DESIRED_TARGET_READJUSTMENT_TIMESPAN,
    INITIAL_TARGET,
    MAX_COINBASE_RANDOM_DATA_SIZE,
    SASHIMI_PER_COIN,
)
from skepticoin.coinstate import CoinState
from skepticoin.consensus import (
    calculate_new_target,
    construct_block_for_mining,
    construct_block_for_mining_genesis,
    construct_minable_summary_genesis,
    construct_minable_summary,
    construct_coinbase_transaction,
    construct_pow_evidence,
    get_block_subsidy,
    get_transaction_fee,
    validate_non_coinbase_transaction_by_itself,
    validate_coinbase_transaction_by_itself,
    validate_block_header_by_itself,
    validate_block_by_itself,
    # validate_non_coinbase_transaction_in_coinstate,
    ValidationError,
    ValidateBlockError,
    ValidateBlockHeaderError,
    ValidateTransactionError,
    ValidatePOWError,
)
from skepticoin.signing import SECP256k1PublicKey, SECP256k1Signature, SignableEquivalent
from skepticoin.datatypes import Transaction, OutputReference, Input, Output, Block, BlockHeader


CHAIN_TESTDATA_PATH = Path(__file__).parent.joinpath("testdata/chain")

example_public_key = SECP256k1PublicKey(b'x' * 64)


def _read_chain_from_disk(max_height):
    coinstate = CoinState.zero()

    for file_path in sorted(CHAIN_TESTDATA_PATH.iterdir()):
        height = int(file_path.name.split("-")[0])
        if height > max_height:
            return coinstate

        block = Block.stream_deserialize(open(file_path, 'rb'))
        coinstate = coinstate.add_block_no_validation(block)

    return coinstate


def get_example_genesis_block():
    # an example (valid) genesis block, but not the one from the actual Skepticoin blockchain.
    return Block.deserialize(computer(
        """000000000000000000000000000000000000000000000000000000000000000000008278968af4bd613aa24a5ccd5280211b3101e3"""
        """ff62621bb11500509d1bbe2a956046240b0100000000000000000000000000000000000000000000000000000000000000000000d7"""
        """38f2c472180cb401f650b12be96ec25bfd9b4e9908c6c9089d9bf26401646f87000000000000000000000000000000000000000000"""
        """0000000000000000000000077a14cfbe21d47f367f23f9a464c765541b1b07bef9f5a95901e0bffe3a1a2f01000100000000000000"""
        """000000000000000000000000000000000000000000000000000000000001000000000001000000012a05f200027878787878787878"""
        """7878787878787878787878787878787878787878787878787878787878787878787878787878787878787878787878787878787878"""
        """787878"""))


def test_calculate_new_target():
    assert calculate_new_target(INITIAL_TARGET, DESIRED_TARGET_READJUSTMENT_TIMESPAN)[:4] == b'\x01\x00\x00\x00'

    assert calculate_new_target(INITIAL_TARGET, DESIRED_TARGET_READJUSTMENT_TIMESPAN * 2)[:4] == b'\x02\x00\x00\x00'
    assert calculate_new_target(INITIAL_TARGET, DESIRED_TARGET_READJUSTMENT_TIMESPAN // 2)[:4] == b'\x00\x80\x00\x00'


def test_construct_minable_summary():
    summary = construct_minable_summary_genesis([
        construct_coinbase_transaction(0, [], immutables.Map(), b"Political statement goes here", example_public_key)
    ], 1231006505, 0)

    summary.serialize()
    # so far... just checking that this doesn't crash :-)


def test_construct_pow_evidence_genesis_block():
    # separate from test_construct_pow_evidence_block_6, because genesis block has no chain sampling (there is no chain)
    coinstate = CoinState.empty()

    transactions = [
        construct_coinbase_transaction(0, [], immutables.Map(), b"Political statement goes here", example_public_key),
    ]

    summary = construct_minable_summary_genesis(transactions, 1231006505, 0)

    evidence = construct_pow_evidence(coinstate, summary, 0, transactions)
    evidence.serialize()
    # no assertions here, just checking that this doesn't crash :-)


def test_construct_pow_evidence_non_genesis_block():
    coinstate = _read_chain_from_disk(5)

    transactions = [
        construct_coinbase_transaction(0, [], immutables.Map(), b"Political statement goes here", example_public_key),
    ]

    summary = construct_minable_summary(coinstate, transactions, 1231006505, 0)

    evidence = construct_pow_evidence(coinstate, summary, coinstate.head().height + 1, transactions)
    evidence.serialize()
    # no assertions here, just checking that this doesn't crash :-)


def test_construct_block_for_mining_no_non_coinbase_transactions():
    coinstate = _read_chain_from_disk(5)

    block = construct_block_for_mining(
        coinstate, [], example_public_key, 1231006505, b'skepticoin is digital bitcoin', 1234)

    assert 6 == block.height
    assert 1 == len(block.transactions)
    assert example_public_key == block.transactions[0].outputs[0].public_key
    assert 1231006505 == block.timestamp
    assert b'skepticoin is digital bitcoin' == block.transactions[0].inputs[0].signature.signature
    assert 1234 == block.nonce


def test_construct_block_for_mining_with_non_coinbase_transactions():
    coinstate = _read_chain_from_disk(5)

    transactions = [Transaction(
        inputs=[Input(
            OutputReference(coinstate.at_head.block_by_height[0].transactions[0].hash(), 0),
            SECP256k1Signature(b'y' * 64),
        )],
        outputs=[Output(9 * SASHIMI_PER_COIN, example_public_key)],
    )]

    block = construct_block_for_mining(
        coinstate, transactions, example_public_key, 1231006505, b'skepticoin is digital bitcoin', 1234)

    assert 6 == block.height
    assert 2 == len(block.transactions)
    assert 11 * SASHIMI_PER_COIN == block.transactions[0].outputs[0].value  # 10 mined + 1 fee
    assert example_public_key == block.transactions[0].outputs[0].public_key
    assert 1231006505 == block.timestamp
    assert b'skepticoin is digital bitcoin' == block.transactions[0].inputs[0].signature.signature
    assert 1234 == block.nonce


def test_validate_non_coinbase_transaction_by_itself_no_inputs():
    transaction = Transaction(
        inputs=[],
        outputs=[Output(30, example_public_key)],
    )

    with pytest.raises(ValidateTransactionError, match=".*No inputs.*"):
        validate_non_coinbase_transaction_by_itself(transaction)


def test_validate_non_coinbase_transaction_by_itself_no_outputs():
    transaction = Transaction(
        inputs=[Input(
            OutputReference(b'a' * 32, 1),
            SECP256k1Signature(b'y' * 64),
        )],
        outputs=[],
    )

    with pytest.raises(ValidateTransactionError, match=".*No outputs.*"):
        validate_non_coinbase_transaction_by_itself(transaction)


def test_validate_non_coinbase_transaction_by_itself_max_size():
    transaction = Transaction(
        inputs=[Input(
            OutputReference(b'a' * 32, 1),
            SECP256k1Signature(b'y' * 64),
        )] * 30_000,
        outputs=[Output(30, example_public_key)]
    )

    with pytest.raises(ValidateTransactionError, match=".*MAX_BLOCK_SIZE.*"):
        validate_non_coinbase_transaction_by_itself(transaction)


def test_validate_non_coinbase_transaction_by_itself_max_total_output():
    transaction = Transaction(
        inputs=[Input(
            OutputReference(b'a' * 32, 1),
            SECP256k1Signature(b'y' * 64),
        )],
        outputs=[Output(21_000_000 * SASHIMI_PER_COIN, example_public_key)]
    )

    with pytest.raises(ValidationError, match=".out of range.*"):
        validate_non_coinbase_transaction_by_itself(transaction)


def test_validate_non_coinbase_transaction_by_itself_no_duplicate_output_references():
    transaction = Transaction(
        inputs=[Input(
            OutputReference(b'a' * 32, 1),
            SECP256k1Signature(b'y' * 64),
        )] * 2,
        outputs=[Output(30, example_public_key)]
    )

    with pytest.raises(ValidateTransactionError, match=".*output_reference referenced more than once.*"):
        validate_non_coinbase_transaction_by_itself(transaction)


def test_validate_non_coinbase_transaction_by_itself_is_not_coinbase():
    transaction = Transaction(
        inputs=[Input(
            OutputReference(b'\x00' * 32, 0),
            SECP256k1Signature(b'y' * 64),
        )],
        outputs=[Output(30, example_public_key)]
    )

    with pytest.raises(ValidateTransactionError, match=".*null-reference in non-coinbase transaction.*"):
        validate_non_coinbase_transaction_by_itself(transaction)


def test_validate_non_coinbase_transaction_by_itself():
    transaction = Transaction(
        inputs=[Input(
            OutputReference(b'a' * 32, 1),
            SignableEquivalent(),
        )],
        outputs=[Output(30, example_public_key)]
    )

    with pytest.raises(ValidateTransactionError, match=".*Non-signature Signature class used.*"):
        validate_non_coinbase_transaction_by_itself(transaction)


def test_validate_signature_for_spend():
    # before anything: create sign_for_spend
    pass  # TODO


def test_get_transaction_fee():
    public_key = SECP256k1PublicKey(b'x' * 64)

    previous_transaction_hash = b'a' * 32

    unspent_transaction_outs = immutables.Map({
        OutputReference(previous_transaction_hash, 0): Output(40, public_key),
        OutputReference(previous_transaction_hash, 1): Output(34, public_key),
    })

    transaction = Transaction(
        inputs=[Input(
            OutputReference(previous_transaction_hash, 1),
            SECP256k1Signature(b'y' * 64),
        )],
        outputs=[Output(30, public_key)]
    )

    assert get_transaction_fee(transaction, unspent_transaction_outs) == 34 - 30  # 4


def test_get_block_fees():
    pass  # TODO


def test_get_block_subsidy():
    assert get_block_subsidy(0) == 10 * SASHIMI_PER_COIN
    assert get_block_subsidy(1_049_999) == 10 * SASHIMI_PER_COIN
    assert get_block_subsidy(1_050_000) == 5 * SASHIMI_PER_COIN
    assert get_block_subsidy(31_499_999) == 1
    assert get_block_subsidy(31_500_000) == 0


def _get_example_coinbase_transaction():
    height = 123
    non_coinbase_transactions = []
    unspent_transaction_outs = immutables.Map()
    miner_public_key = b'x' * 32
    random_data = b"No need to get all political here"
    return construct_coinbase_transaction(
        height, non_coinbase_transactions, unspent_transaction_outs, random_data, miner_public_key)


def test_validate_coinbase_transaction_by_itself_for_valid_coinbase():
    cb = _get_example_coinbase_transaction()
    validate_coinbase_transaction_by_itself(cb)


def test_validate_coinbase_transaction_by_itself_exactly_1_input():
    cb = _get_example_coinbase_transaction()
    cb.inputs.append(deepcopy(cb.inputs[0]))

    with pytest.raises(ValidateTransactionError, match=".*1 input.*"):
        validate_coinbase_transaction_by_itself(cb)


def test_validate_coinbase_transaction_by_itself_reference_real_output():
    cb = _get_example_coinbase_transaction()
    cb.inputs[0].output_reference = OutputReference(b'c' * 32, 4)

    with pytest.raises(ValidateTransactionError, match=".*must create.*thin air.*"):
        validate_coinbase_transaction_by_itself(cb)


def test_validate_coinbase_transaction_by_itself_should_have_coinbasedata():
    cb = _get_example_coinbase_transaction()
    cb.inputs[0].signature = SECP256k1Signature(b'c' * 64)

    with pytest.raises(ValidateTransactionError, match=".*CoinbaseData.*"):
        validate_coinbase_transaction_by_itself(cb)


def test_validate_coinbase_transaction_by_itself_maximum_coinbasedata_size():
    cb = _get_example_coinbase_transaction()
    cb.inputs[0].signature.signature = b'x' * (MAX_COINBASE_RANDOM_DATA_SIZE + 1)

    with pytest.raises(ValidateTransactionError, match=".*MAX_COINBASE_RANDOM_DATA_SIZE.*"):
        validate_coinbase_transaction_by_itself(cb)


def test_validate_block_header_by_itself_no_errors():
    validate_block_header_by_itself(get_example_genesis_block().header, current_timestamp=1615209942)


def test_validate_block_header_by_itself_for_bad_pow():
    block = get_example_genesis_block()
    block.header.summary.nonce = 1
    with pytest.raises(ValidatePOWError):
        validate_block_header_by_itself(block.header, 1615209942)


def test_validate_block_header_by_itself_for_block_from_the_future():
    block = get_example_genesis_block()
    with pytest.raises(ValidateBlockHeaderError):
        validate_block_header_by_itself(block.header, block.header.summary.timestamp - 100)


def test_validate_block_by_itself_for_correct_block():
    validate_block_by_itself(get_example_genesis_block(), 1615209942)


def test_validate_block_by_itself_no_transactions():
    block = get_example_genesis_block()
    block.transactions = []

    with pytest.raises(ValidateBlockError, match=".*No transactions.*"):
        validate_block_by_itself(block, 1615209942)


def test_validate_block_by_itself_max_block_size(mocker):
    # work around get_block_fees... because we would need a coinstate w/ 50_000 unspent outputs to make it work.
    mocker.patch("skepticoin.consensus.get_block_fees", return_value=0)

    transactions = [
        Transaction(
            inputs=[Input(
                OutputReference(i.to_bytes(32, 'big'), 1),
                SECP256k1Signature(b'y' * 64),
            )],
            outputs=[Output(1, example_public_key)]
        ) for i in range(50_000)
    ]

    block = construct_block_for_mining_genesis(transactions, example_public_key, 1231006505, b'', 22)

    with pytest.raises(ValidateBlockError, match=".*MAX_BLOCK_SIZE.*"):
        validate_block_by_itself(block, 1615209942)


def test_validate_block_by_itself_invalid_coinbase_transaction():
    coinstate = CoinState.empty()

    transactions = [Transaction(
        inputs=[Input(
            OutputReference(b'x' * 32, 1),
            SECP256k1Signature(b'y' * 64),
        )],
        outputs=[Output(1, example_public_key)]
    )]

    summary = construct_minable_summary(coinstate, transactions, 1615209942, 39)
    evidence = construct_pow_evidence(coinstate, summary, 0, transactions)
    block = Block(BlockHeader(summary, evidence), transactions)

    with pytest.raises(ValidateTransactionError, match=".*thin air.*"):
        validate_block_by_itself(block, 1615209942)


def test_validate_block_by_itself_for_mismatched_heights():
    block = get_example_genesis_block()
    block.height = 1

    with pytest.raises(ValidateBlockError, match=".*height.*"):
        validate_block_by_itself(block, 1615209942)


def test_validate_block_by_itself_invalid_non_coinbase_transaction():
    pass


def test_validate_block_by_itself_no_duplicate_transactions():
    pass


def test_validate_block_by_itself_no_duplicate_output_references():
    pass


def test_validate_block_by_itself_incorrect_merkle_root_hash():
    pass


def test_validate_non_coinbase_transaction_in_coinstate_invalid_output_reference():
    # I started on the below, but the amount of setup is getting excessive... perhaps it's going to be easier to
    # express these tests when more mechanisms of _creation_ are available? we'll see
    '''
    previous_transaction_hash = b'a' * 32

    unspent_transaction_outs = immutables.Map()

    transaction = Transaction(
        inputs=[Input(
            OutputReference(previous_transaction_hash, 1),
            SECP256k1Signature(b'y' * 64),
        )],
        outputs=[Output(30, public_key)]
    )

    with pytest.raises(ValidateTransactionError, match=r".*does not exist.*") as e:
        validate_non_coinbase_transaction_in_coinstate

    '''


def test_validate_non_coinbase_transaction_in_coinstate_invalid_signature():
    pass  # TODO


def test_validate_non_coinbase_transaction_in_coinstate_overspending():
    pass  # TODO


def test_validate_non_coinbase_transaction_in_coinstate_valid():
    pass  # TODO



from datetime import datetime

from .utils import (
    open_or_init_wallet,
    create_chain_dir,
    read_chain_from_disk,
    configure_logging_from_args,
    DefaultArgumentParser,
)

from ..params import SASHIMI_PER_COIN


def main():
    parser = DefaultArgumentParser()
    args = parser.parse_args()
    configure_logging_from_args(args)

    create_chain_dir()
    coinstate = read_chain_from_disk()
    wallet = open_or_init_wallet()
    print(
        wallet.get_balance(coinstate) / SASHIMI_PER_COIN, "SKEPTI at h. %s," % coinstate.head().height,
        datetime.fromtimestamp(coinstate.head().timestamp).isoformat())



import os
from pathlib import Path
from decimal import Decimal
from datetime import datetime
import random
from skepticoin.datatypes import Block
from skepticoin.networking.threading import NetworkingThread
from skepticoin.coinstate import CoinState
from typing import Any, Dict

from skepticoin.params import SASHIMI_PER_COIN
from skepticoin.consensus import construct_block_for_mining
from skepticoin.signing import SECP256k1PublicKey
from skepticoin.wallet import Wallet, save_wallet
from skepticoin.utils import block_filename
from skepticoin.cheating import MAX_KNOWN_HASH_HEIGHT
from time import time
from multiprocessing import Process, Lock, Queue, synchronize
from .utils import (
    initialize_peers_file,
    create_chain_dir,
    read_chain_from_disk,
    open_or_init_wallet,
    start_networking_peer_in_background,
    check_for_fresh_chain,
    configure_logging_from_args,
    DefaultArgumentParser,
)


def run_miner(args: Any, wallet_lock: synchronize.Lock, queue: Queue, miner_id: int) -> None:
    miner = Miner(args, wallet_lock, queue, miner_id)
    miner()


class Miner:
    def __init__(self, args: Any, wallet_lock: synchronize.Lock, queue: Queue, miner_id: int) -> None:
        self.args = args
        self.wallet_lock = wallet_lock
        self.queue = queue
        self.miner_id = miner_id
        self.wallet: Wallet
        self.coinstate: CoinState
        self.thread: NetworkingThread

    def send_message(self, message_type: str, data: Any) -> None:
        message = (self.miner_id, message_type, data)
        self.queue.put(message)

    def prepare(self):
        configure_logging_from_args(self.args)

        create_chain_dir()
        self.coinstate = read_chain_from_disk()

        self.wallet_lock.acquire()
        self.wallet = open_or_init_wallet()
        self.wallet_lock.release()

        initialize_peers_file()
        self.thread = start_networking_peer_in_background(self.args, self.coinstate)
        self.thread.local_peer.show_stats()

        if check_for_fresh_chain(self.thread):
            self.thread.local_peer.show_stats()

        if self.thread.local_peer.chain_manager.coinstate.head().height <= MAX_KNOWN_HASH_HEIGHT:
            self.send_message("info", "Your blockchain is not just old, it is ancient; ABORTING")
            os.exit(0)

    def get_key_for_mined_block(self) -> bytes:
        self.wallet_lock.acquire()
        public_key = self.wallet.get_annotated_public_key("reserved for potentially mined block")
        save_wallet(self.wallet)
        self.wallet_lock.release()
        return public_key

    def get_balance(self) -> Decimal:
        self.wallet_lock.acquire()
        balance = self.wallet.get_balance(self.coinstate) / Decimal(SASHIMI_PER_COIN)
        self.wallet_lock.release()
        return balance

    def handle_mined_block(self, block: Block) -> None:
        self.coinstate = self.coinstate.add_block(block, int(time()))
        with open(Path('chain') / block_filename(block), 'wb') as f:
            f.write(block.serialize())

        self.send_message("found_block", block_filename(block))
        self.send_message("balance", self.get_balance())

        self.thread.local_peer.chain_manager.set_coinstate(self.coinstate)
        self.thread.local_peer.network_manager.broadcast_block(block)

    def __call__(self) -> None:
        self.prepare()
        self.send_message("start_balance", self.get_balance())
        self.send_message("info", "Starting mining: A repeat minter")

        try:
            public_key = self.get_key_for_mined_block()
            nonce = random.randrange(1 << 32)
            last_round_second = int(time())
            hashes = 0

            while True:
                current_second = int(time())

                if current_second > last_round_second:
                    self.send_message("hashes", (current_second, hashes))
                    last_round_second = current_second
                    hashes = 0

                self.coinstate, transactions = self.thread.local_peer.chain_manager.get_state()
                increasing_time = max(current_second, self.coinstate.head().timestamp + 1)

                block = construct_block_for_mining(
                    self.coinstate, transactions, SECP256k1PublicKey(public_key), increasing_time, b'', nonce)

                hashes += 1
                nonce = (nonce + 1) % (1 << 32)

                if block.hash() < block.target:
                    self.handle_mined_block(block)
                    public_key = self.get_key_for_mined_block()

        except KeyboardInterrupt:
            self.send_message("info", "KeyboardInterrupt")

        finally:
            self.send_message("info", "Stopping networking thread")
            self.thread.stop()

            self.send_message("info", "Waiting for networking thread to stop")
            self.thread.join()

            self.send_message("info", "Done; waiting for Python-exit")


def main() -> None:
    parser = DefaultArgumentParser()
    parser.add_argument('-n', default=1, type=int, help='number of miner instances')
    args = parser.parse_args()

    queue = Queue()
    wallet_lock = Lock()
    processes = []

    for i in range(args.n):
        if i > 0:
            args.dont_listen = True

        process = Process(target=run_miner, daemon=True, args=(args, wallet_lock, queue, i))
        process.start()
        processes.append(process)

    hash_stats: Dict[int, Dict[int, int]] = {}
    start_time = datetime.now()

    balance = 0
    start_balance = 0

    try:
        while True:
            miner_id, message_type, data = queue.get()

            if message_type == "hashes":
                timestamp, hashes = data

                if timestamp not in hash_stats:
                    hash_stats[timestamp] = {}

                hash_stats[timestamp][miner_id] = hashes

                current_time = int(time())

                # delete old hashing stats
                for ts, miner_stats in list(hash_stats.items()):
                    if ts < current_time and len(miner_stats) < args.n:
                        print(f"WARNING: Only got stats from {len(miner_stats)} "
                              f"of {args.n} processes at timestamp {timestamp}.")
                        del hash_stats[ts]

                if len(hash_stats[timestamp]) == args.n:
                    now = datetime.now()
                    now_str = now.strftime("%Y-%m-%d %H:%M:%S")
                    uptime = now - start_time
                    uptime_str = str(uptime).split(".")[0]

                    mined = balance - start_balance
                    total_hashes = sum(hash_stats[timestamp].values())
                    del hash_stats[timestamp]

                    mine_speed = (float(mined) / uptime.total_seconds()) * 60 * 60
                    print(f"{now_str} | uptime: {uptime_str} | {total_hashes:>3} hash/sec" +
                          f" | mined: {mined:>3} SKEPTI | {mine_speed:5.2f} SKEPTI/h")

            elif message_type == "balance":
                balance = data

            elif message_type == "start_balance":
                start_balance = data
                balance = data

            elif message_type == "found_block":
                print(f"miner {miner_id:2} found block: {data}")

            elif message_type == "info":
                print(f"miner {miner_id:2}: {data}")

            else:
                print(f"unhandled message_type {message_type} from {miner_id}, data={data}")

    except KeyboardInterrupt:
        pass
    finally:
        for process in processes:
            process.join()



import traceback

from .utils import (
    initialize_peers_file,
    create_chain_dir,
    read_chain_from_disk,
    open_or_init_wallet,
    start_networking_peer_in_background,
    configure_logging_from_args,
    DefaultArgumentParser,
)


class EverythingIsNone:
    def __getattr__(self, attr):
        return None


def main():
    parser = DefaultArgumentParser()
    parser.add_argument("script_file")
    args = parser.parse_args()
    configure_logging_from_args(args)

    create_chain_dir()
    coinstate = read_chain_from_disk()
    wallet = open_or_init_wallet()
    initialize_peers_file()
    thread = start_networking_peer_in_background(args, coinstate)
    thread.local_peer.show_stats()

    try:
        globals = {
            'thread': thread,
            'wallet': wallet,
            'get_coinstate': lambda: thread.local_peer.chain_manager.coinstate,
        }
        try:
            exec(open(args.script_file, 'r').read(), globals)
        except BaseException:
            print(traceback.format_exc())

    finally:
        print("Stopping networking thread")
        thread.stop()
        print("Waiting for networking thread to stop")
        thread.join()
        print("Done; waiting for Python-exit")



"""
The general theme of these tests is: let's do some end-to-end tests, so we at least hit a lot of code paths and expose
the most obvious of mistakes.
"""

import pytest
from time import time
import logging
from pathlib import Path
from time import sleep
import socket

from skepticoin.signing import SignableEquivalent, SECP256k1PublicKey
from skepticoin.datatypes import Block, Transaction, Input, Output, OutputReference
from skepticoin.coinstate import CoinState
from skepticoin.networking.messages import InventoryMessage
from skepticoin.networking.threading import NetworkingThread
from skepticoin.networking.utils import load_peers_from_list

CHAIN_TESTDATA_PATH = Path(__file__).parent.joinpath("../testdata/chain")


class FakeDiskInterface:
    def save_block(self, block):
        pass

    def update_peer_db(self, remote_peer):
        pass

    def save_transaction_for_debugging(self, transaction):
        pass


def _read_chain_from_disk(max_height):
    coinstate = CoinState.zero()

    for file_path in sorted(CHAIN_TESTDATA_PATH.iterdir()):
        height = int(file_path.name.split("-")[0])
        if height > max_height:
            return coinstate

        block = Block.stream_deserialize(open(file_path, 'rb'))
        coinstate = coinstate.add_block_no_validation(block)

    return coinstate


def _try_to_connect(host, port):
    # just a blocking connect to check that a server is up and listening, then close the connection.
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect_ex((host, port))
    sock.close()


def test_ibd_integration(caplog):
    # just testing the basics: if we set up 2 threads, one with a coinstate, and one without, will the chain propagate?

    caplog.set_level(logging.INFO)

    coinstate = _read_chain_from_disk(5)

    thread_a = NetworkingThread(coinstate, 12412, FakeDiskInterface())
    thread_a.start()

    _try_to_connect('127.0.0.1', 12412)

    thread_b = NetworkingThread(CoinState.zero(), 12413, FakeDiskInterface())
    thread_b.local_peer.network_manager.disconnected_peers = load_peers_from_list([('127.0.0.1', 12412, "OUTGOING")])
    thread_b.start()

    try:
        start_time = time()
        while True:
            if thread_b.local_peer.chain_manager.coinstate.head().height == 5:
                break

            if time() > start_time + 5:
                print("\n".join(str(r) for r in caplog.records))
                raise Exception("IBD failed")

            sleep(0.01)

    finally:
        thread_a.stop()
        thread_a.join()

        thread_b.stop()
        thread_b.join()


def test_broadcast_transaction(caplog, mocker):
    # just testing the basics: is a broadcast transaction stored in the transaction pool on the other side?

    # By turning off transaction-validation, we can use an invalid transaction in this test.
    mocker.patch("skepticoin.networking.peer.validate_non_coinbase_transaction_by_itself")
    mocker.patch("skepticoin.networking.peer.validate_non_coinbase_transaction_in_coinstate")

    caplog.set_level(logging.INFO)

    coinstate = _read_chain_from_disk(5)

    thread_a = NetworkingThread(coinstate, 12412, FakeDiskInterface())
    thread_a.start()

    _try_to_connect('127.0.0.1', 12412)

    thread_b = NetworkingThread(coinstate, 12413, FakeDiskInterface())
    thread_b.local_peer.network_manager.disconnected_peers = load_peers_from_list([('127.0.0.1', 12412, "OUTGOING")])
    thread_b.start()

    previous_hash = coinstate.at_head.block_by_height[0].transactions[0].hash()

    # Not actually a valid transaction (not signed)
    transaction = Transaction(
        inputs=[Input(OutputReference(previous_hash, 0), SignableEquivalent())],
        outputs=[Output(10, SECP256k1PublicKey(b'x' * 64))],
    )

    try:
        # give both peers some time to find each other
        start_time = time()
        while True:
            if (len(thread_a.local_peer.network_manager.get_active_peers()) > 0 and
                    len(thread_b.local_peer.network_manager.get_active_peers()) > 0):
                break

            if time() > start_time + 5:
                print("\n".join(str(r) for r in caplog.records))
                raise Exception("Peers can't connect")

            sleep(0.01)

        # broadcast_transaction... the part that we're testing
        thread_a.local_peer.network_manager.broadcast_transaction(transaction)

        # wait until it's picked up on the other side
        start_time = time()
        while True:
            if len(thread_b.local_peer.chain_manager.transaction_pool) > 0:
                break

            if time() > start_time + 5:
                print("\n".join(str(r) for r in caplog.records))
                raise Exception("Transaction broadcast failed")

            sleep(0.01)

    finally:
        thread_a.stop()
        thread_a.join()

        thread_b.stop()
        thread_b.join()


@pytest.mark.skip(reason="fickle test")
def test_broadcast_message_closed_connection_handling(caplog, mocker):
    caplog.set_level(logging.INFO)

    coinstate = _read_chain_from_disk(5)

    thread_a = NetworkingThread(coinstate, 12412, FakeDiskInterface())
    thread_a.start()

    _try_to_connect('127.0.0.1', 12412)

    thread_b = NetworkingThread(coinstate, 12413, FakeDiskInterface())
    thread_b.local_peer.network_manager.disconnected_peers = load_peers_from_list([('127.0.0.1', 12412, "OUTGOING")])
    thread_b.start()

    try:
        # give both peers some time to find each other
        start_time = time()
        while True:
            if (len(thread_a.local_peer.network_manager.get_active_peers()) > 0 and
                    len(thread_b.local_peer.network_manager.get_active_peers()) > 0):
                break

            if time() > start_time + 5:
                print("\n".join(str(r) for r in caplog.records))
                raise Exception("Peers can't connect")

            sleep(0.01)

        # do a hard disconnect (without going through local_peer.disconnect)
        for remote_peer in thread_a.local_peer.network_manager.get_active_peers():
            remote_peer.sock.close()

        # broadcast a message
        thread_a.local_peer.network_manager.broadcast_message(InventoryMessage([]))

        # the error should have been logged
        assert len([r for r in caplog.records if 'ChainManager.broadcast_message' in str(r)]) > 0

    finally:
        thread_a.stop()
        thread_a.join()

        thread_b.stop()
        thread_b.join()
